﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Globalization;
using System.Net;
using System.Security.Permissions;
using System.Text.RegularExpressions;
using System.Web;
using System.Web.UI;
using System.Web.UI.WebControls.WebParts;
using System.Xml;
using Microsoft.SharePoint;


namespace Mazin.SPGoogleMapPart
{
    [ToolboxItem(false)]
    [AspNetHostingPermission(SecurityAction.Demand, Level = AspNetHostingPermissionLevel.Minimal)]
    [AspNetHostingPermission(SecurityAction.InheritanceDemand, Level = AspNetHostingPermissionLevel.Minimal)]
    public class VisualWebPart : WebPart
    {
        // Visual Studio might automatically update this path when you change the Visual Web Part project item.
        private const string _ascxPath = @"~/_CONTROLTEMPLATES/SPGoogleMapPart/MapWebPartUserControl.ascx";
        private readonly List<Marker> _markers = new List<Marker>();
        private string _addressColumnName;

        // map settings
        private double _centreLatitude;
        private double _centreLongitude;
        private int _dataMode;
        private string _descrColumnName;
        private int _height = 300;
        private string _latitudeColumnName;
        private string _listItemIdParamName;
        private string _listQuery;
        private string _listUrl;
        private string _longitudeColumnName;
        private string _mapType = "HYBRID";

        // provider that provides data while connected to other webpart
        private IWebPartTable _provider;
        private string _titleColumnName;
        private string _webUrl;
        private int _width = 400;
        private int _zoom = 1;

        // points (Google calls them markers) from datasource to be displayed on map

        [Category("Map"), Personalizable(PersonalizationScope.Shared),
         WebDescription("Latitude coordinate of the map centre"), WebBrowsable(false),
         WebDisplayName("Map Centre Latitude")]
        public double CentreLatitude
        {
            get { return _centreLatitude; }
            set { _centreLatitude = value; }
        }

        [WebDisplayName("Map Centre Longitude"), Personalizable(PersonalizationScope.Shared), WebBrowsable(false),
         WebDescription("Longitude coordinate of the map centre"), Category("Map")]
        public double CentreLongitude
        {
            get { return _centreLongitude; }
            set { _centreLongitude = value; }
        }

        [WebDescription("Map Height"), Personalizable(PersonalizationScope.Shared), Category("Map"), WebBrowsable(false)
        , WebDisplayName("Map Height")]
        public int MapHeight
        {
            get { return _height; }
            set { _height = value; }
        }

        [WebDescription("Map Type"), Personalizable(PersonalizationScope.Shared), WebBrowsable(false),
         WebDisplayName("Map type"), Category("Map")]
        public string MapType
        {
            get { return _mapType; }
            set { _mapType = value; }
        }

        [WebBrowsable(false), WebDisplayName("Map width"), Category("Map"), WebDescription("Map Width"),
         Personalizable(PersonalizationScope.Shared)]
        public int MapWidth
        {
            get { return _width; }
            set { _width = value; }
        }

        [WebDescription("Zoom Level"), Personalizable(PersonalizationScope.Shared), WebBrowsable(false),
         WebDisplayName("Zoom level"), Category("Map")]
        public int ZoomLevel
        {
            get { return _zoom; }
            set { _zoom = value; }
        }

        [WebDescription("Name of address column"), Personalizable(PersonalizationScope.Shared), WebBrowsable(false),
         WebDisplayName("Address column"), Category("Connection")]
        public string AddressColumn
        {
            get { return _addressColumnName; }
            set { _addressColumnName = value; }
        }

        [WebDescription("Name of longitude column"), Personalizable(PersonalizationScope.Shared), WebBrowsable(false),
         WebDisplayName("Longitude column"), Category("Connection")]
        public string LongitudeColumn
        {
            get { return _longitudeColumnName; }
            set { _longitudeColumnName = value; }
        }

        [WebDescription("Name of latitude column"), Personalizable(PersonalizationScope.Shared), WebBrowsable(false),
         WebDisplayName("Latitude column"), Category("Connection")]
        public string LatitudeColumn
        {
            get { return _latitudeColumnName; }
            set { _latitudeColumnName = value; }
        }

        [WebDescription("Name of description column"), Personalizable(PersonalizationScope.Shared), WebBrowsable(false),
         WebDisplayName("Description column"), Category("Connection")]
        public string DescriptionColumn
        {
            get { return _descrColumnName; }
            set { _descrColumnName = value; }
        }

        [WebDescription("Name of title column"), Personalizable(PersonalizationScope.Shared), WebBrowsable(false),
         WebDisplayName("Title column"), Category("Connection")]
        public string TitleColumn
        {
            get { return _titleColumnName; }
            set { _titleColumnName = value; }
        }

        [WebDescription("Url of Source list"), Personalizable(PersonalizationScope.Shared), WebBrowsable(false),
         WebDisplayName("List URL"), Category("Connection")]
        public string ListUrl
        {
            get { return _listUrl; }
            set { _listUrl = value; }
        }

        [WebDescription("Url of Source list"), Personalizable(PersonalizationScope.Shared), WebBrowsable(false),
         WebDisplayName("Web URL"), Category("Connection")]
        public string WebUrl
        {
            get { return _webUrl; }
            set { _webUrl = value; }
        }

        [WebDescription("CAML query applied on the source list"), Personalizable(PersonalizationScope.Shared),
         WebBrowsable(false), WebDisplayName("List Query"), Category("Connection")]
        public string ListQuery
        {
            get { return _listQuery; }
            set { _listQuery = value; }
        }

        [WebDescription("Name of URL Id parameter for item to be displayed"),
         Personalizable(PersonalizationScope.Shared), WebBrowsable(false), WebDisplayName("List item Id parameter name")
        , Category("Connection")]
        public string ListItemIdParamName
        {
            get { return _listItemIdParamName; }
            set { _listItemIdParamName = value; }
        }

        [WebDescription("Mode to get data"), Personalizable(PersonalizationScope.Shared), WebBrowsable(false),
         WebDisplayName("Data mode"), Category("Connection")]
        public int DataMode
        {
            get { return _dataMode; }
            set { _dataMode = value; }
        }

        /// <summary>
        ///   IWebPartTable consumer method.
        /// </summary>
        /// <param name = "provider">IWebPartTable provider.</param>
        [ConnectionConsumer("Table")]
        public void GetTableProvider(IWebPartTable provider)
        {
            _provider = provider;
        }


        /// <summary>
        ///   Method uses configured data source to get list of markers to be displayed
        /// </summary>
        private void GetData()
        {
            DataTable table = null;

            if (_provider != null)
            {
                DataView data = null;
                _provider.GetTableData((d) =>
                                           {
                                               data = (DataView) d;
                                               table = data.Table;
                                           });
                if (data == null) return;
            }
            else
            {
                try
                {
                    SPSite site = SPContext.Current.Site;
                    using (SPWeb web = site.OpenWeb(_webUrl))
                    {
                        SPList list = web.GetList(_listUrl);
                        SPQuery query = new SPQuery();
                        query.Query = _listQuery;
                        if (!String.IsNullOrEmpty(_listItemIdParamName) &&
                            Page.Request.Params[_listItemIdParamName] != null)
                        {
                            string itemId = Page.Request.Params[_listItemIdParamName];
                            query.Query =
                                String.Format(
                                    "<Query><Where><Eq><FieldRef Name=\"ID\" /><Value Type=\"Counter\">{0}</Value></Eq></Where></Query>",
                                    itemId);
                        }
                        SPListItemCollection items = list.GetItems(query);
                        table = items.GetDataTable();
                    }
                }
                catch
                {
                }
            }

            if (table != null &&
                (!String.IsNullOrEmpty(_addressColumnName) ||
                 (!String.IsNullOrEmpty(_latitudeColumnName) && !String.IsNullOrEmpty(_longitudeColumnName))))
            {
                List<Marker> tmps = new List<Marker>();
                for (int i = 0; i < table.Rows.Count; i++)
                {
                    DataRow row = table.Rows[i];
                    try
                    {
                        Marker tmp = new Marker();

                        if (!String.IsNullOrEmpty(_latitudeColumnName) && !String.IsNullOrEmpty(_longitudeColumnName))
                        {
                            FillCoordinates(row, tmp);
                        }

                        if (!String.IsNullOrEmpty(_addressColumnName) && tmp.Latitude == 0 && tmp.Longitude == 0)
                        {
                            tmp.Address = (string) row[_addressColumnName];
                            ResolveAddress(tmp);
                        }

                        tmp.Title = (string) row[_titleColumnName];
                        if (!String.IsNullOrEmpty(_descrColumnName)) tmp.Description = row[_descrColumnName].ToString();

                        if (tmp.Latitude != 0.0 && tmp.Longitude != 0.0) tmps.Add(tmp);
                    }
                    catch
                    {
                    }
                }
                if (tmps.Count > 0) _markers.AddRange(tmps);
            }
        }

        /// <summary>
        ///   Fill marker's coordinates from source data row
        /// </summary>
        private void FillCoordinates(DataRow row, Marker marker)
        {
            try
            {
                double dPos = 0;
                bool lngFound = false;
                string strPos = row[_latitudeColumnName].ToString();
                bool parseOk = Double.TryParse(strPos, out dPos);
                if (!parseOk)
                {
                    Regex r =
                        new Regex(
                            @"(?<first>(((?<degrees>\d+)\xb0(?<minutes>\d+)′(?<seconds>\d+)″)(?<quadrant>[NSEW]))|((?<degrees>\d+\x2E\d+)(?<quadrant>[NSEW]))|((?<sign>[-]){0,1}(?<degrees>\d+\x2E\d+))|((?<sign>[-]){0,1}(?<degrees>\d+)\xb0\s(?<minutes>\d+\x2E\d+))|(((?<degrees>\d+)d\s(?<minutes>\d+)′\s(?<seconds>\d+)″\s)(?<quadrant>[NSEW])))((,{0,1}\s)(?<lng>(((?<degrees>\d+)\xb0(?<minutes>\d+)′(?<seconds>\d+)″)(?<quadrant>[NSEW]))|((?<degrees>\d+\x2E\d+)(?<quadrant>[NSEW]))|((?<sign>[-]){0,1}(?<degrees>\d+\x2E\d+))|((?<sign>[-]){0,1}(?<degrees>\d+)\xb0\s(?<minutes>\d+\x2E\d+))|(((?<degrees>\d+)d\s(?<minutes>\d+)′\s(?<seconds>\d+)″\s)(?<quadrant>[NSEW]))){0,1})",
                            RegexOptions.Singleline | RegexOptions.ExplicitCapture);
                    Match match = r.Match(strPos);
                    if (match.Success)
                    {
                        string sign = match.Groups["sign"].Success ? match.Groups["sign"].Captures[0].Value : "";
                        string quadrant = match.Groups["quadrant"].Success
                                              ? match.Groups["quadrant"].Captures[0].Value
                                              : "";
                        string degrees = match.Groups["degrees"].Captures[0].Value;
                        string minutes = match.Groups["minutes"].Success
                                             ? match.Groups["minutes"].Captures[0].Value
                                             : "0";
                        string seconds = match.Groups["seconds"].Success
                                             ? match.Groups["seconds"].Captures[0].Value
                                             : "0";
                        marker.Latitude = ConstructCoordinate(sign, quadrant, degrees, minutes, seconds);
                        if (match.Groups["lng"] != null)
                        {
                            lngFound = true;
                            sign = match.Groups["sign"].Success ? match.Groups["sign"].Captures[1].Value : "";
                            quadrant = match.Groups["quadrant"].Success
                                           ? match.Groups["quadrant"].Captures[1].Value
                                           : "";
                            degrees = match.Groups["degrees"].Captures[1].Value;
                            minutes = match.Groups["minutes"].Success ? match.Groups["minutes"].Captures[1].Value : "0";
                            seconds = match.Groups["seconds"].Success ? match.Groups["seconds"].Captures[1].Value : "0";
                            marker.Longitude = ConstructCoordinate(sign, quadrant, degrees, minutes, seconds);
                        }
                    }
                }
                else marker.Latitude = dPos;
                if (!lngFound)
                {
                    strPos = row[_longitudeColumnName].ToString();
                    parseOk = Double.TryParse(strPos, out dPos);
                    if (!parseOk)
                    {
                        Regex r =
                            new Regex(
                                @"(?<first>(((?<degrees>\d+)\xb0(?<minutes>\d+)′(?<seconds>\d+)″)(?<quadrant>[NSEW]))|((?<degrees>\d+\x2E\d+)(?<quadrant>[NSEW]))|((?<sign>[-]){0,1}(?<degrees>\d+\x2E\d+))|((?<sign>[-]){0,1}(?<degrees>\d+)\xb0\s(?<minutes>\d+\x2E\d+))|(((?<degrees>\d+)d\s(?<minutes>\d+)′\s(?<seconds>\d+)″\s)(?<quadrant>[NSEW])))((,{0,1}\s)(?<lng>(((?<degrees>\d+)\xb0(?<minutes>\d+)′(?<seconds>\d+)″)(?<quadrant>[NSEW]))|((?<degrees>\d+\x2E\d+)(?<quadrant>[NSEW]))|((?<sign>[-]){0,1}(?<degrees>\d+\x2E\d+))|((?<sign>[-]){0,1}(?<degrees>\d+)\xb0\s(?<minutes>\d+\x2E\d+))|(((?<degrees>\d+)d\s(?<minutes>\d+)′\s(?<seconds>\d+)″\s)(?<quadrant>[NSEW]))){0,1})",
                                RegexOptions.Singleline | RegexOptions.ExplicitCapture);
                        Match match = r.Match(strPos);
                        if (match.Success)
                        {
                            string sign = match.Groups["sign"].Success ? match.Groups["sign"].Captures[0].Value : "";
                            string quadrant = match.Groups["quadrant"].Success
                                                  ? match.Groups["quadrant"].Captures[0].Value
                                                  : "";
                            string degrees = match.Groups["degrees"].Captures[0].Value;
                            string minutes = match.Groups["minutes"].Success
                                                 ? match.Groups["minutes"].Captures[0].Value
                                                 : "0";
                            string seconds = match.Groups["seconds"].Success
                                                 ? match.Groups["seconds"].Captures[0].Value
                                                 : "0";
                            marker.Longitude = ConstructCoordinate(sign, quadrant, degrees, minutes, seconds);
                        }
                    }
                    else marker.Longitude = dPos;
                }
            }
            catch
            {
            }
        }

        /// <summary>
        ///   Construct decimal representation used by Google from structured form of coordinate
        /// </summary>
        private static double ConstructCoordinate(string sign, string quadrant, string degrees, string minutes,
                                                  string seconds)
        {
            double retValue = 0;
            double d = Double.Parse(degrees);
            double m = Double.Parse(minutes);
            double s = Double.Parse(seconds);
            retValue = d + m/60 + s/3600;
            if (sign == "-" || quadrant == "S" || quadrant == "W") retValue *= -1;
            return retValue;
        }

        /// <summary>
        ///   This method resolve markers address to coordinates using Google Geocode API.
        /// </summary>
        
        private static void ResolveAddress(Marker marker)
       {

           ServicePointManager.ServerCertificateValidationCallback = delegate { return true; };
           WebClient wc = new WebClient();
            string result =
              wc.DownloadString(
                  String.Format("https://maps.googleapis.com/maps/api/geocode/xml?sensor=false&address=" + HttpUtility.UrlEncode(marker.Address)));

           XmlDocument xmlDocument = new XmlDocument();
           xmlDocument.LoadXml(result);
           
           XmlNode status = xmlDocument.SelectSingleNode("//GeocodeResponse/status");

          
         if (status != null && status.InnerText == "OK")
          {
               marker.Latitude =
                  Double.Parse(
                       xmlDocument.SelectSingleNode("/GeocodeResponse/result/geometry/location/lat").InnerText,
                       CultureInfo.InvariantCulture);
              marker.Longitude =
                  Double.Parse(
                      xmlDocument.SelectSingleNode("/GeocodeResponse/result/geometry/location/lng").InnerText,
                      CultureInfo.InvariantCulture);
        }
}







        public override EditorPartCollection CreateEditorParts()
        {
            ArrayList editorParts = new ArrayList();
            MapEditorPart editor = new MapEditorPart();
            editor.ID = ID + "_editorPart";
            editorParts.Add(editor);
            return new EditorPartCollection(editorParts);
        }

        protected override void CreateChildControls()
        {
            Control control = Page.LoadControl(_ascxPath);
            GetData();
            VisualWebPartUserControl childControls = ((VisualWebPartUserControl) control);
            childControls.Markers = _markers.ToArray();
            childControls.Zoom = _zoom;
            childControls.MapWidth = _width;
            childControls.MapHeight = _height;
            if (_centreLongitude == 0 && _centreLatitude == 0)
                CalculateCentre(out _centreLatitude, out _centreLongitude);
            childControls.CentreLatitude = _centreLatitude;
            childControls.CentreLongitude = _centreLongitude;
            childControls.MapType = _mapType;
            Controls.Add(control);
        }

        /// <summary>
        ///   When center point is not configured, this method is used to calculate it using found markers
        /// </summary>
        private void CalculateCentre(out double cLat, out double cLng)
        {
            double maxLat = -200, minLat = 200;
            double maxLng = -200, minLng = 200;

            foreach (Marker marker in _markers)
            {
                if (maxLat < marker.Latitude) maxLat = marker.Latitude;
                if (minLat > marker.Latitude) minLat = marker.Latitude;
                if (maxLng < marker.Longitude) maxLng = marker.Longitude;
                if (minLng > marker.Longitude) minLng = marker.Longitude;
            }

            cLat = (maxLat + minLat)/2;
            cLng = (maxLng + minLng)/2;
        }
    }
}